/**
	 *The Program may contain sample source code or programs, which illustrate programming techniques. These samples are found in the Samples Gallery.  You may copy, modify, and distribute these samples, or their modifications, in any form, internally or as part of your application or related documentation. These samples have not been tested under all conditions and are provided to you by IBM without obligation of support of any kind.  IBM PROVIDES THESE SAMPLES  "AS IS", SUBJECT TO ANY STATUTORY WARRANTIES THAT CANNOT BE EXCLUDED. IBM MAKES NO WARRANTIES OR CONDITIONS, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OR CONDITIONS OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT REGARDING THESE SAMPLES OR TECHNICAL SUPPORT, IF ANY.
	 */

import java.lang.reflect.InvocationTargetException;
import java.text.MessageFormat;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.uml2.ConnectableElement;
import org.eclipse.uml2.EventOccurrence;
import org.eclipse.uml2.Feature;
import org.eclipse.uml2.Lifeline;
import org.eclipse.uml2.Message;
import org.eclipse.uml2.MessageEnd;
import org.eclipse.uml2.Property;
import org.eclipse.uml2.Type;
import org.eclipse.uml2.TypedElement;
import org.eclipse.uml2.UML2Package;

import com.ibm.xtools.emf.msl.ResourceSetReadOperation;
import com.ibm.xtools.modeler.UMLModeler;
import com.ibm.xtools.notation.ElementBridge;
import com.ibm.xtools.notation.ModelBridge;
import com.ibm.xtools.notation.View;
import com.ibm.xtools.pluglets.Pluglet;

public class FanInMetric extends Pluglet {

	/**
	 * Walks the selected elements and logs the fan-in metric
	 * of each visited class elements.
	 */
	public void plugletmain(String[] args) {

		// Get selection
		List elements = UMLModeler.getUMLUIHelper().getSelectedElements();

		// For each selected element
		boolean logged = false;

		try {
			for (Iterator iter = elements.iterator(); iter.hasNext();) {
				Object object = iter.next();
				
				// If a view, try to get semantic bridge
				if (object instanceof View) {
					ModelBridge bridge = ((View)object).getModelBridge();
					if (bridge instanceof ElementBridge) {
						object = ((ElementBridge)bridge).getElement();
					}
				}

				// If a class
				if (object instanceof org.eclipse.uml2.Class) {
					org.eclipse.uml2.Class clazz = (org.eclipse.uml2.Class) object;

					// Calculate fan-in
					int fanIn = calcFanIn(clazz);

					// Log it
					String message = MessageFormat.format(
						"{0} fan-in is: {1}", //$NON-NLS-1$
						new Object[] {clazz.getName(), String.valueOf(fanIn)});

					logged = true;

					out.println(message);
				}
			}
		} catch (InterruptedException e) {
			out.println("The operation was interrupted"); //$NON-NLS-1$

		} catch (InvocationTargetException e1) {
			out.println("An unexpected exception prevented the operation from completing successfully.  See Error Log View for details."); //$NON-NLS-1$

			throw new RuntimeException(e1);
		}

		if (false == logged) {
			out.println("Could not calculate fan-in for selection"); //$NON-NLS-1$
		}

	}

	/**
	 * Calculates the fan-in metric of the selected element
	 * 
	 * @param clazz
	 *            The class to calculate the fan-in for
	 * @return The fan-in metric
	 * @throws InterruptedException
	 *             The operation was cancelled
	 * @throws InvocationTargetException
	 *             An exception occured that prevented fan-in calculation
	 *  
	 */
	private int calcFanIn(final org.eclipse.uml2.Class clazz)
		throws InvocationTargetException, InterruptedException {

		final int size[] = {0};

		// Model access must be wrapped within a ResourceSetReadOperation
		UMLModeler.getEditingDomain().run(new ResourceSetReadOperation() {

			protected void execute(IProgressMonitor monitor)
				throws InvocationTargetException, InterruptedException {

				// Set that holds the classes that reference the specified class
				HashSet referencingClasses = new HashSet();

				// Handle dependencies that can be inferred from operations
				getRefencingClassesByOperation(clazz, referencingClasses);

				// Handle dependencies that can be inferred from properties
				getRefencingClassesByProperty(clazz, referencingClasses);

				// Make sure not to count the specified class
				referencingClasses.remove(clazz);

				// Fan-in metric is the number of classes that reference the
				// specified class
				size[0] = referencingClasses.size();
			}
		}, new NullProgressMonitor());

		return size[0];
	}

	/**
	 * Adds to the set specified by <code>referencingClasses</code> all
	 * classes that reference the class specified by <code>clazz</code>
	 * through operations
	 * 
	 * @param clazz
	 *            The class to look for references to
	 * @param referencingClasses
	 *            The set of classes that reference the specified class
	 */
	private void getRefencingClassesByOperation(org.eclipse.uml2.Class clazz,
			HashSet referencingClasses) {

		// For each of the specified class' features
		EList features = clazz.getFeatures();
		for (Iterator iter = features.iterator(); iter.hasNext();) {
			Feature feature = (Feature) iter.next();

			// Look for referencer to the feature using their Signature feature
			EReference[] eReferences = new EReference[] {UML2Package.eINSTANCE
				.getMessage_Signature()};
			Collection messages = UMLModeler.getEObjectHelper().getReferencers(
				feature, eReferences);

			// For each referencer
			for (Iterator iterRef = messages.iterator(); iterRef.hasNext();) {

				// Make sure it is a Message
				Object object = iterRef.next();
				if (object instanceof Message) {

					Message message = (Message) object;

					// Get sender message end and add the class at the other end
					// to the set
					MessageEnd sendMessageEnd = message.getSendEvent();
					addOtherClassAtMessageEnd(sendMessageEnd,
						referencingClasses);
				}
			}
		}
	}

	/**
	 * Retrieve the class representing the lifeline covered by the MessageEnd
	 * specified by <code>messageEnd</code>, and add to the Set specified by
	 * <code>set</code>
	 * 
	 * @param messageEnd
	 *            MessageEnd to get class from
	 * @param set
	 *            Set to add the class to
	 */
	private void addOtherClassAtMessageEnd(MessageEnd messageEnd, HashSet set) {

		// try to cast as an EventOccurence
		if (messageEnd instanceof EventOccurrence) {
			EventOccurrence eventOccurence = (EventOccurrence) messageEnd;

			// There should be only one lifeline
			EList lifelines = eventOccurence.getCovereds();
			if (lifelines.size() == 1) {
				Lifeline lifeline = (Lifeline) lifelines.get(0);
				ConnectableElement connectableElement = lifeline
					.getRepresents();

				// ConnectableElement should be a property
				if (connectableElement instanceof Property) {
					Property property = (Property) connectableElement;

					// Type of property determines caller class
					Type type = property.getType();
					if (type instanceof org.eclipse.uml2.Class) {
						set.add(type);
					}
				}
			}
		}
	}

	/**
	 * Adds to the set specified by <code>referencingClasses</code> all
	 * classes that reference the class specified by <code>clazz</code>
	 * through a property
	 * 
	 * @param clazz
	 *            The class to look for references to
	 * @param referencingClasses
	 *            The set of classes that reference the specified class
	 */
	private void getRefencingClassesByProperty(org.eclipse.uml2.Class clazz,
			HashSet referencingClasses) {

		// Look for referencer to the class as a type
		EReference[] eReferences = new EReference[] {UML2Package.eINSTANCE
			.getTypedElement_Type()};

		Collection typedElements = UMLModeler.getEObjectHelper()
			.getReferencers(clazz, eReferences);

		// For each typed elements referencing the specified class
		for (Iterator iter = typedElements.iterator(); iter.hasNext();) {

			TypedElement typedElement = (TypedElement) iter.next();

			// Check if it is a Property
			if (typedElement instanceof Property) {

				Property property = (Property) typedElement;

				// Add class that owns the Property to the set
				org.eclipse.uml2.Class ownerClass = property.getClass_();

				if (null != ownerClass) {
					referencingClasses.add(ownerClass);
				}
			}
		}
	}

}
